/*
 GB 64MB Mighty Flash Cart - Backup save slots using GBxCart RW
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 7/11/2016
 Last Modified: 17/12/2018
 
 A once click program to create a folder with the current date/time and backup your save slots with your given save slot
 names to that folder.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("insideGadgets - 64MB Mighty Flash Cart - Backup Save Slots\n");
	printf("##########################################################\n");
	
	read_config();
	
	// Open COM port
	if (com_test_port() == 0) {
		printf("Device not connected and couldn't be auto detected\n");
		read_one_letter();
		return 1;
	}
	printf("Connected on COM port: %i\n", cport_nr+1);
	
	// Break out of any existing functions on ATmega
	set_mode('0');
	
	
	// Get cartridge mode - Gameboy or Gameboy Advance
	cartridgeMode = request_value(CART_MODE);
	
	// Get firmware version
	gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
	
	// Get PCB version
	gbxcartPcbVersion = request_value(READ_PCB_VERSION);
	
	// Check if in GB mode
	if (gbxcartPcbVersion == PCB_1_1 && cartridgeMode == GBA_MODE) {
		printf("\nYou must be in GB Mode, 5V\n");
		read_one_letter();
		return 0;
	}
	
	// Set 5V
	if (gbxcartPcbVersion == PCB_1_3) {
		set_mode(VOLTAGE_5V);
		delay_ms(500);
	}
	
	
	// Get the date/time to be used for the folder
	time_t rawtime;
	struct tm* timeinfo;
	char timebuffer[25];
	
	time (&rawtime);
	timeinfo = localtime (&rawtime);
	strftime (timebuffer, 80, "%Y-%m-%d_%H-%M-%S", timeinfo);
	
	// Create folder
	char mkfolder[250];
	strncpy(mkfolder, "mkdir ", 7);
	strncat(mkfolder, timebuffer, 25);
	system(mkfolder);
	
	
	printf("\nBacking up save slots to folder: %s\n", timebuffer);
	
	// Loop 8 slots
	for (uint8_t slot = 1; slot <= 8; slot++) {
		// Read save slot name if given
		uint16_t bank = ((0x20000 * slot) + 0x8000) / 0x4000;
		set_bank(0x2100, bank & 0xFF);
		
		currAddr = 0x4000;
		endAddr = 0x7FFF;
		set_number(currAddr, SET_START_ADDRESS);
		
		set_mode(READ_ROM_RAM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop(); // Stop reading ROM (as we will bank switch)
		
		// Blank out game title
		uint8_t gameTitleUsed = 0;
		for (uint8_t b = 0; b < 8; b++) {
			gameTitle[b] = 0;
		}
		// Read save name and check for non-printable text
		for (uint16_t addr = 0; addr <= 8; addr++) {
			char headerChar = readBuffer[addr];
			if ((headerChar >= 0x30 && headerChar <= 0x39) || // 0-9
				 (headerChar >= 0x41 && headerChar <= 0x5A) || // A-Z
				 (headerChar >= 0x61 && headerChar <= 0x7A) || // a-z
				 (headerChar >= 0x24 && headerChar <= 0x29) || // #$%&'()
				 (headerChar == 0x2E) || // .
				 (headerChar == 0x5F) || // _
				 (headerChar == 0x20)) { // Space
				gameTitle[addr] = headerChar;
				gameTitleUsed = 1;
			}
			// Replace with an underscore
			else if (headerChar == 0x3A) { // : 
				gameTitle[addr] = '_';
			}
			else {
				gameTitle[addr] = '\0';
				break;
			}
		}
		printf ("Save Slot %i, Name: %s\n", slot, gameTitle);
		
		
		// Save slot to file
		char saveSlotNumber[2];
		snprintf(saveSlotNumber, 2, "%d", slot);
		
		char titleFilename[250];
		strncpy(titleFilename, timebuffer, 25);
		strncat(titleFilename, "/Save_Slot_", 12);
		strncat(titleFilename, saveSlotNumber, 2);
		if (gameTitleUsed == 1) {
			strncat(titleFilename, "_", 2);
			strncat(titleFilename, gameTitle, 9);
		}
		strncat(titleFilename, ".sav", 5);
		//printf("%s\n", titleFilename);
		
		
		// Read the save slot
		FILE *romFile = fopen(titleFilename, "wb");
		bank -= 2;
		for (uint8_t x = 0; x < 2; x++) {				
			set_bank(0x2100, (bank+x) & 0xFF);
			currAddr = 0x4000;
			
			// Set start address and rom reading mode
			set_number(currAddr, SET_START_ADDRESS);
			set_mode(READ_ROM_RAM);
			
			// Read data
			while (currAddr < endAddr) {
				com_read_bytes(romFile, 64);
				currAddr += 64;
				
				// Request 64 bytes more
				if (currAddr < endAddr) {
					com_read_cont();
				}
			}
			com_read_stop(); // Stop reading ROM
		}
		fclose(romFile);
	}
	
	printf("Finished");
	
	return 0;
}