/*
 GB 64MB Mighty Flash Cart - Multi-Game Cart Maker (Backend)
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 17/12/2018
 Last Modified: 17/12/2018
 
 This is used for creating the output.gb file from the GUI.
 
 Argc1, 0 = Use file name or 1 = Use game title
 Argc2, 0 = No save slots, 1 = 8 save slots
 Argc3+ = Files
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <libgen.h>

#ifdef _WIN32
#include <Windows.h>
#include <direct.h>
#define GetCurrentDir _getcwd
#else
#include <unistd.h>
#define GetCurrentDir getcwd
#endif

// Read one letter from stdin
char read_one_letter (void) {
	char c = getchar();
	while (getchar() != '\n' && getchar() != EOF);
	return c;
}

int main(int argc, char **argv) {
	
	printf("insideGadgets - 64MB Mighty Cart - Multi-Game Maker v1.0\n");
	printf("########################################################\n");

	uint8_t buffer[20] = {0x00};
	uint8_t writeBuffer[1] = {0x00};
	uint16_t romBank = 2;
	uint8_t useGameTitle = 0;
	uint8_t useSaveSlots = 0;
	uint8_t saveFileSlots = 8;
	
	if (argc >= 3) {
		// Use game title?
		if (atoi(argv[1]) == 1) {
			useGameTitle = 1;
		}
		// Use save slots
		if (atoi(argv[2]) >= 1) {
			useSaveSlots = 1;
		}
		
		// Copy the base rom file to the output file
		system("copy base.gb output.gb > /nul");
		
		FILE *romWriteFile = fopen("output.gb", "r+b");
		
		// Pad to cover the loader and save slots if save slots are used
		if (useSaveSlots == 1) {
			fseek(romWriteFile, 0x8000, SEEK_SET);
			
			// Makes 128KB for loader
			uint32_t remainingBytes = 6143;
			for (uint8_t x = 0; x < 16; x++) {
				buffer[x] = 0xFF;
			}
			for (uint32_t x = 0; x <= remainingBytes; x++) {
				fwrite(buffer, 1, 16, romWriteFile);
			}
			
			// Pad the save slots (128KB x 8 slots)
			remainingBytes = (8192 * saveFileSlots) - 1;
			for (uint8_t x = 0; x < 16; x++) {
				buffer[x] = 0x00;
			}
			for (uint32_t x = 0; x <= remainingBytes; x++) {
				fwrite(buffer, 1, 16, romWriteFile);
			}
			
			// Change starting game slot
			romBank = 8 + (saveFileSlots * 8);
		}
		
		// Write if we are using saves
		if (useSaveSlots == 1) {
			fseek(romWriteFile, 0x32FE, SEEK_SET);
			buffer[0] = 8;
			fwrite(buffer, 1, 1, romWriteFile);
		}
		
		// Write how many games there are
		fseek(romWriteFile, 0x32FF, SEEK_SET);
		uint8_t gamesCount = argc;
		buffer[0] = gamesCount - 3;
		fwrite(buffer, 1, 1, romWriteFile);
		
		// Write the game names
		for (uint8_t g = 3; g < gamesCount; g++) {
			// Read ROM game title
			uint16_t startLocation = 0x0134;
			char gameTitle[18];
			FILE *romReadFile = fopen(argv[g], "rb");
			fseek(romReadFile, startLocation, SEEK_SET);
			fread(buffer, 1, 18, romReadFile);
			
			// Clear game title
			for (uint8_t x = 0; x < 18; x++) {
				gameTitle[x] = 0;
			}
			
			// Parse game title
			for (uint8_t x = 0; x < 17; x++) {
				if ((buffer[x] >= 0x30 && buffer[x] <= 0x39) || // 0-9
					 (buffer[x] >= 0x41 && buffer[x] <= 0x5A) || // A-Z
					 (buffer[x] >= 0x61 && buffer[x] <= 0x7A) || // a-z
					 (buffer[x] >= 0x24 && buffer[x] <= 0x29) || // #$%&'()
					 (buffer[x] == 0x2E) || // .
					 (buffer[x] == 0x5F) || // _
					 (buffer[x] == 0x20)) { // Space
					gameTitle[x] = buffer[x];
				}
				// Replace with an underscore
				else if (buffer[x] == 0x3A) { // :
					gameTitle[x] = '_';
				}
				else {
					gameTitle[x] = 0;
				}
			}
			gameTitle[17] = 0;
			
			// MBC type
			startLocation = 0x0147;
			fseek(romReadFile, startLocation, SEEK_SET);
			fread(buffer, 1, 1, romReadFile);
			uint8_t mbcType = buffer[0];
			
			// ROM size
			startLocation = 0x0148;
			fseek(romReadFile, startLocation, SEEK_SET);
			fread(buffer, 1, 1, romReadFile);
			uint8_t romSize = buffer[0];
			
			// File size
			fseek (romReadFile, 0, SEEK_END);
			long fileSize = ftell (romReadFile);
			fclose(romReadFile);
			
			// Write ROM bank up to
			buffer[0] = romBank & 0xFF;
			fwrite(buffer, 1, 1, romWriteFile);
			
			// Write the MBC type
			if (mbcType <= 0x0D && romSize <= 0x04) { // MBC1 <= 512KB ROM
				writeBuffer[0] = 0xA0;
			}
			else if (mbcType <= 0x0D) { // MBC1 >= 1MB ROM
				writeBuffer[0] = 0xC0;
			}
			else if (mbcType >= 0x0E && mbcType <= 0x18) { // MBC3
				writeBuffer[0] = 0xE0;
			}
			else { // MBC5
				writeBuffer[0] = 0xF0;
			}
			writeBuffer[0] |= (romBank >> 8); // Set 4 high bits of romBank
			fwrite(writeBuffer, 1, 1, romWriteFile);
			
			// Write the game title or filename
			if (useGameTitle == 1) {
				fwrite(gameTitle, 1, 18, romWriteFile);
			}
			else {
				// Clear game title
				for (uint8_t x = 0; x < 18; x++) {
					gameTitle[x] = 0;
				}
				
				char tempTitle[200];
				char *dotLocation;
				strncpy(tempTitle, basename(argv[g]), 200);
				dotLocation = strrchr(tempTitle, '.'); // Check for the last dot in the file name
				
				// Dot found
				if (dotLocation-tempTitle > 1 && dotLocation-tempTitle < 18) {
					//printf("%s %i\n", tempTitle, dotLocation-tempTitle);
					strncpy(gameTitle, tempTitle, dotLocation-tempTitle);
				}
				else if (dotLocation-tempTitle > 18) {
					strncpy(gameTitle, tempTitle, 18);
				}
				else {
					strncpy(gameTitle, tempTitle, 18);
				}
				fwrite(gameTitle, 1, 18, romWriteFile);
			}
			gameTitle[18] = 0;
			
			printf("Adding: %s\n", gameTitle);
			
			// Calculate ROM banks used
			romBank += fileSize / 16384;
		}
		fclose(romWriteFile);
		
		// Append the ROM files
		printf("Appending files");
		for (uint8_t g = 3; g < gamesCount; g++) {
			char appendCommand[255];
			strncpy(appendCommand, "copy /b output.gb+\"", 21);
			strncat(appendCommand, argv[g], 200);
			strncat(appendCommand, "\" output.gb", 12);
			strncat(appendCommand, " > /nul", 8);
			
			system(appendCommand);
			
			printf(".");
		}
		
		//printf("Total file size: %i\n", fileSizeTotal);
	}
	
	//read_one_letter();
	
	return 0;
}